import { getContextItem } from "../lib/context.ts";
import { Doodler, Vector } from "@bearmetal/doodler";
import { Spline, TrackSegment, TrackSystem } from "../track/system.ts";
import { Debuggable } from "../lib/debuggable.ts";
import { lerp, lerpAngle, map } from "../math/lerp.ts";

export class Train extends Debuggable {
  nodes: Vector[] = [];

  cars: TrainCar[] = [];

  path: Spline<TrackSegment>;
  t: number;

  spacing = 0;

  speed = 5;

  aabb!: AABB;

  get segments() {
    return Array.from(
      new Set(this.cars.flatMap((c) => c.segments.values().toArray())),
    );
  }

  constructor(track: Spline<TrackSegment>, cars: TrainCar[], t = 0) {
    super("train", "path");
    this.path = track;
    this.path.pointSpacing = 4;
    this.cars = cars;
    this.t = this.cars.reduce((acc, c) => acc + c.length, 0) +
      (this.cars.length - 1) * this.spacing;
    this.t = this.t / this.path.pointSpacing;

    let currentOffset = 0;
    // try {
    for (const car of this.cars) {
      car.train = this;
      currentOffset += car.moveAlongPath(this.t - currentOffset, true) +
        this.spacing / this.path.pointSpacing;
    }
    // } catch {
    //   currentOffset = 0;
    //   console.log("Reversed");
    //   for (const car of this.cars.toReversed()) {
    //     for (const [i, bogie] of car.bogies.entries().toArray().reverse()) {
    //       currentOffset += bogie.length;
    //       const a = this.path.followEvenPoints(this.t - currentOffset);
    //       car.setBogiePosition(a.p, i);
    //       this.nodes.push(a.p);
    //       car.segments.add(a.segmentId);
    //     }
    //   }
    // }

    this.updateAABB();
  }

  move(dTime: number) {
    if (!this.speed) return;
    this.t = this.t + (this.speed / this.path.pointSpacing) * dTime * 10;
    // % this.path.evenPoints.length; // This should probably be on the track system
    let currentOffset = 0;
    for (const car of this.cars) {
      // This needs to be moved to the car itself
      // if (!car.points) return;
      // const [a, b] = car.points;
      // const nA = this.path.followEvenPoints(this.t - currentOffset);
      // a.set(nA.p);
      // currentOffset += car.length;
      // const nB = this.path.followEvenPoints(this.t - currentOffset);
      // b.set(nB.p);
      // currentOffset += this.spacing;
      // car.segments = [nA.segmentId, nB.segmentId];
      // car.draw();

      currentOffset += car.moveAlongPath(this.t - currentOffset) +
        (this.spacing / this.path.pointSpacing);
    }
    // this.draw();
    this.updateAABB();
  }

  updateAABB() {
    const minX = Math.min(...this.cars.map((c) => c.aabb.x));
    const maxX = Math.max(...this.cars.map((c) => c.aabb.x + c.aabb.width));
    const minY = Math.min(...this.cars.map((c) => c.aabb.y));
    const maxY = Math.max(...this.cars.map((c) => c.aabb.y + c.aabb.height));
    this.aabb = {
      pos: new Vector(minX, minY),
      x: minX,
      y: minY,
      width: maxX - minX,
      height: maxY - minY,
      center: new Vector(minX, minY).add(
        new Vector(maxX - minX, maxY - minY).div(2),
      ),
    };
  }

  // draw() {
  //   const doodler = getContextItem<Doodler>("doodler");
  //   this.path.draw();
  //   for (const [i, node] of this.nodes.entries()) {
  //     // doodler.drawCircle(node, 10, { color: "purple", weight: 3 });
  //     doodler.fillCircle(node, 2, { color: "purple" });
  //     // const next = this.nodes[i + 1];
  //     // if (next) {
  //     //   const to = Vector.sub(node.point, next.point);
  //     //   to.setMag(40);
  //     //   doodler.line(next.point, Vector.add(to, next.point))
  //     // }
  //   }
  // }

  draw() {
    for (const car of this.cars) {
      car.draw();
    }
  }

  override debugDraw(): void {
    const debug = getContextItem<Debug>("debug");
    const doodler = getContextItem<Doodler>("doodler");
    if (debug.path) {
      // doodler.drawLine(this.path.points, {
      //   color: "red",
      //   weight: 3,
      // });
      const colors = getContextItem<string[]>("colors");
      for (const [i, p] of this.path.evenPoints.entries()) {
        const color = colors[
          Math.floor(
            map(i, 0, this.path.evenPoints.length, 0, colors.length),
          )
        ];
        doodler.drawCircle(p.p, 2, { color, weight: .5 });
      }
    }

    if (debug.train) {
      const track = getContextItem<TrackSystem>("track");
      const colors = getContextItem<string[]>("colors").slice();
      colors.push(colors.shift()!);
      colors.push(colors.shift()!);
      colors.push(colors.shift()!);
      for (const [i, segmentId] of this.segments.entries().toArray()) {
        const segment = track.getSegment(segmentId);
        segment &&
          doodler.drawBezier(...segment.points, {
            color: colors[i % colors.length],
            weight: 3,
          });
      }
    }
    if (debug.aabb) {
      doodler.deferDrawing(() => {
        doodler.drawRect(this.aabb.pos, this.aabb.width, this.aabb.height, {
          color: "orange",
        });
        doodler.drawCircle(this.aabb.center, 2, {
          color: "lime",
        });
      });
    }
  }

  real2Track(length: number) {
    return length / this.path.pointSpacing;
  }
}

interface Bogie {
  pos: Vector;
  angle: number;
  length: number;
  sprite?: ISprite & { offset?: Vector };
  rotate?: boolean;
}

export class TrainCar extends Debuggable {
  img: HTMLImageElement;
  imgWidth: number;
  imgHeight: number;
  sprite?: ISprite;

  points?: [Vector, Vector, ...Vector[]];
  _length: number;
  leading: number = 0;

  bogies: Bogie[] = [];

  segments: Set<string> = new Set();

  train?: Train;

  aabb!: AABB;

  constructor(
    length: number,
    trailing: number,
    img: HTMLImageElement,
    w: number,
    h: number,
    sprite?: ISprite,
  ) {
    super(true, "car", "bogies", "angles");
    this.img = img;
    this.sprite = sprite;
    this.imgWidth = w;
    this.imgHeight = h;
    this._length = length;

    this.bogies = [
      {
        pos: new Vector(0, 0),
        angle: 0,
        length: length,
      },
      {
        pos: new Vector(0, 0),
        angle: 0,
        length: trailing,
      },
    ];

    this.updateAABB();
  }

  get length() {
    return this.bogies.reduce((acc, b) => acc + b.length, 0) + this.leading;
  }

  setBogiePosition(pos: Vector, idx: number) {
    this.bogies[idx].pos.set(pos);
  }

  update(dTime: number, t: number) {
    if (this.train) {
      for (const [i, bogie] of this.bogies.entries()) {
        const a = this.train.path.followEvenPoints(t - this._length * i);
      }
    }
  }

  updateAABB() {
    let minX = Infinity;
    let maxX = -Infinity;
    let minY = Infinity;
    let maxY = -Infinity;

    this.bogies.forEach((bogie, index) => {
      // Unit vector in the direction the bogie is facing.
      const u = new Vector(Math.cos(bogie.angle), Math.sin(bogie.angle));
      // Perpendicular vector (to thicken the rectangle).
      const v = new Vector(-Math.sin(bogie.angle), Math.cos(bogie.angle));

      // For the first bogie, extend in the opposite direction by this.leading.
      let front = bogie.pos.copy();
      if (index === 0) {
        front = front.sub(u.copy().rotate(Math.PI).mult(this.leading));
      }
      // Rear point is at bogie.pos plus the bogie length.
      const rear = bogie.pos.copy().add(
        u.copy().rotate(Math.PI).mult(bogie.length),
      );

      // Calculate half the height to offset from the center line.
      const halfHeight = this.imgHeight / 2;

      // Calculate the four corners of the rectangle.
      const corners = [
        front.copy().add(v.copy().mult(halfHeight)),
        front.copy().add(v.copy().mult(-halfHeight)),
        rear.copy().add(v.copy().mult(halfHeight)),
        rear.copy().add(v.copy().mult(-halfHeight)),
      ];

      // Update the overall AABB limits.
      corners.forEach((corner) => {
        minX = Math.min(minX, corner.x);
        minY = Math.min(minY, corner.y);
        maxX = Math.max(maxX, corner.x);
        maxY = Math.max(maxY, corner.y);
      });
    });
    this.aabb = {
      pos: new Vector(minX, minY),
      center: new Vector(minX, minY).add(
        new Vector(maxX - minX, maxY - minY).div(2),
      ),
      x: minX,
      y: minY,
      width: maxX - minX,
      height: maxY - minY,
    };
  }

  moveAlongPath(t: number, initial = false): number {
    if (!this.train) return 0;
    let offset = this.leading / this.train.path.pointSpacing;
    this.segments.clear();
    for (const [i, bogie] of this.bogies.entries()) {
      const a = this.train.path.followEvenPoints(t - offset);
      a.tangent.rotate(TWO_PI);
      offset += bogie.length / this.train.path.pointSpacing;
      this.setBogiePosition(a.p, i);
      if (initial) bogie.angle = a.tangent.heading();
      else {
        bogie.angle = lerpAngle(a.tangent.heading(), bogie.angle, .1);
      }
      this.segments.add(a.segmentId);
    }
    this.updateAABB();
    return offset;
  }

  draw() {
    const doodler = getContextItem<Doodler>("doodler");
    const [a, b] = this.bogies;
    const origin = Vector.add(Vector.sub(a.pos, b.pos).div(2), b.pos);
    const angle = Vector.sub(b.pos, a.pos).heading();

    doodler.drawCircle(origin, 4, { color: "blue" });

    doodler.drawRotated(origin, angle, () => {
      this.sprite
        ? doodler.drawSprite(
          this.img,
          this.sprite.at,
          this.sprite.width,
          this.sprite.height,
          origin.copy().sub(this.imgWidth / 2, this.imgHeight / 2),
          this.imgWidth,
          this.imgHeight,
        )
        : doodler.drawImage(
          this.img,
          origin.copy().sub(this.imgWidth / 2, this.imgHeight / 2),
        );
    });
  }
  override debugDraw(...args: unknown[]): void {
    const doodler = getContextItem<Doodler>("doodler");
    const debug = getContextItem<Debug>("debug");
    if (debug.bogies) {
      doodler.deferDrawing(() => {
        for (const [i, b] of this.bogies.entries()) {
          const next = this.bogies[i + 1];
          if (!next) continue;
          const dist = Vector.dist(b.pos, next.pos);
          doodler.drawCircle(b.pos, 5, { color: "red" });
          doodler.fillText(
            dist.toFixed(1).toString(),
            b.pos.copy().add(10, 10),
            100,
            {
              color: "white",
            },
          );
        }
      });
    }
    if (debug.car) {
      doodler.deferDrawing(() => {
        doodler.drawLine(this.bogies.map((b) => b.pos), {
          color: "blue",
          weight: 2,
        });
        doodler.deferDrawing(() => {
          const colors = getContextItem<string[]>("colors");
          for (const [i, b] of this.bogies.entries()) {
            doodler.drawCircle(b.pos, 5, { color: colors[i % colors.length] });
            doodler.fillText(
              b.length.toString(),
              b.pos.copy().add(10, 0),
              100,
              {
                color: "white",
              },
            );
          }
        });
      });

      if (debug.aabb) {
        doodler.deferDrawing(() => {
          doodler.drawRect(this.aabb.pos, this.aabb.width, this.aabb.height, {
            color: "white",
            weight: .5,
          });
          doodler.drawCircle(this.aabb.center, 2, {
            color: "yellow",
          });
          doodler.fillText(
            this.aabb.width.toFixed(1).toString(),
            this.aabb.center.copy().add(10, 10),
            100,
            {
              color: "white",
            },
          );
        });
      }
    }

    if (debug.angles) {
      doodler.deferDrawing(() => {
        const ps: { pos: Vector; angle: number }[] = [];
        for (const [i, b] of this.bogies.entries()) {
          ps.push({ pos: b.pos, angle: b.angle });
          const next = this.bogies[i + 1];
          if (!next) continue;
          const heading = Vector.sub(next.pos, b.pos);
          const p = b.pos.copy().add(heading.mult(.5));
          ps.push({ pos: p, angle: heading.heading() });
        }
        for (const p of ps) {
          doodler.dot(p.pos, { color: "green" });
          doodler.fillText(
            p.angle.toFixed(2).toString(),
            p.pos.copy().add(0, 20),
            100,
            {
              color: "white",
            },
          );
        }
      });
    }
  }
}

interface ISprite {
  at: Vector;
  width: number;
  height: number;
}
